# #############################################################################
# __init__.py
# ===========
# Author : Matthieu Simeoni [matthieu.simeoni@gmail.com]
# #############################################################################

r"""
Routines for building graphs from tessellations/point sets in :math:`\mathbb{R}^3`.
"""

import numpy as np
import scipy.linalg as linalg
import scipy.sparse as sp
import scipy.sparse.linalg as splinalg
import scipy.spatial as spatial
from pygsp.graphs import Graph
import healpy as hp
from typing import Tuple


def cvxhull_graph(R: np.ndarray, cheb_normalized: bool = True, compute_differential_operator: bool = True) \
        -> Tuple[Graph, float]:
    r"""
    Build the convex hull graph of a point set in :math:`\mathbb{R}^3`.

    The graph edges have exponential-decay weighting.

    Definitions of the graph Laplacians:

    .. math::

        L     = I - D^{-1/2} W D^{-1/2},\qquad        L_{n} = (2 / \mu_{\max}) L - I

    Parameters
    ----------
    R : :py:class:`~numpy.ndarray`
        (N,3) Cartesian coordinates of point set with size N. All points must be **distinct**.
    cheb_normalized : bool
        Rescale Laplacian spectrum to [-1, 1].
    compute_differential_operator : bool
        Computes the graph gradient.

    Returns
    -------
    G : :py:class:`~pygsp.graphs.Graph`
        If ``cheb_normalized = True``, ``G.Ln`` is created (Chebyshev Laplacian :math:`L_{n}` above)
        If ``compute_differential_operator = True``, ``G.D`` is created and contains the gradient.
    rho : float
        Scale parameter :math:`\rho` corresponding to the average distance of a point
        on the graph to its nearest neighbors.

    Examples
    --------

    .. plot::

        import numpy as np
        from pycgsp.graph import cvxhull_graph
        from pygsp.plotting import plot_graph
        theta, phi = np.linspace(0,np.pi,6, endpoint=False)[1:], np.linspace(0,2*np.pi,9, endpoint=False)
        theta, phi = np.meshgrid(theta, phi)
        x,y,z = np.cos(phi)*np.sin(theta), np.sin(phi)*np.sin(theta), np.cos(theta)
        R = np.stack((x.flatten(), y.flatten(), z.flatten()), axis=-1)
        G, _ = cvxhull_graph(R)
        plot_graph(G)

    Warnings
    --------
    In the newest version of PyGSP (> 0.5.1) the convention is changed: ``Graph.D`` is the divergence operator and
    ``Graph.D.transpose()`` the gradient (see routine `Graph.compute_differential_operator <https://pygsp.readthedocs.io/en/latest/reference/graphs.html#pygsp.graphs.Graph.compute_differential_operator>`_). The code should be adapted when this new version is released.

    """

    # Form convex hull to extract nearest neighbors. Each row in
    # cvx_hull.simplices is a triangle of connected points.
    cvx_hull = spatial.ConvexHull(R)
    cols = np.roll(cvx_hull.simplices, shift=1, axis=-1).reshape(-1)
    rows = cvx_hull.simplices.reshape(-1)

    # Form sparse affinity matrix from extracted pairs
    W = sp.coo_matrix((cols * 0 + 1, (rows, cols)),
                      shape=(cvx_hull.vertices.size, cvx_hull.vertices.size))
    # Symmetrize the matrix to obtain an undirected graph.
    extended_row = np.concatenate([W.row, W.col])
    extended_col = np.concatenate([W.col, W.row])
    W.row, W.col = extended_row, extended_col
    W.data = np.concatenate([W.data, W.data])
    W = W.tocsr().tocoo()  # Delete potential duplicate pairs

    # Weight matrix elements according to the exponential kernel
    distance = linalg.norm(cvx_hull.points[W.row, :] -
                           cvx_hull.points[W.col, :], axis=-1)
    rho = np.mean(distance)
    W.data = np.exp(- (distance / rho) ** 2)
    W = W.tocsc()

    G = _graph_laplacian(W, R, compute_differential_operator=compute_differential_operator,
                         cheb_normalized=cheb_normalized)
    return G, rho


def _graph_laplacian(W, R, compute_differential_operator=False, cheb_normalized=False):
    # Form Graph Laplacian
    G = Graph(W, gtype='undirected', lap_type='normalized', coords=R)
    G.compute_laplacian()  # Stored in G.L, sparse matrix, csc ordering
    if compute_differential_operator is True:
        G.compute_differential_operator()  # stored in G.D, also accessible via G.grad() or G.div() (for the adjoint).
    else:
        pass

    if cheb_normalized:
        D_max = splinalg.eigsh(G.L, k=1, return_eigenvectors=False)
        Ln = (2 / D_max[0]) * G.L - sp.identity(W.shape[0], dtype=np.float, format='csc')
        G.Ln = Ln
    else:
        pass
    return G


def healpix_nngraph(nside: int, cheb_normalized: bool = True, compute_differential_operator: bool = True) \
        -> Tuple[Graph, float]:
    r"""
    Build the nearest neighbour graph of a HEALPix spherical point set.

    The graph edges have exponential-decay weighting.

    Definitions of the graph Laplacians:

    .. math::

        L     = I - D^{-1/2} W D^{-1/2},\qquad        L_{n} = (2 / \mu_{\max}) L - I

    Parameters
    ----------
    nside: int
        Parameter NSIDE of the `HEALPix discretisation scheme <https://healpix.jpl.nasa.gov/>`_.
    cheb_normalized : bool
        Rescale Laplacian spectrum to [-1, 1].
    compute_differential_operator : bool
        Computes the graph gradient.

    Returns
    -------
    G : :py:class:`~pygsp.graphs.Graph`
        If ``cheb_normalized = True``, ``G.Ln`` is created (Chebyshev Laplacian :math:`L_{n}` above)
        If ``compute_differential_operator = True``, ``G.D`` is created and contains the gradient.
    rho : float
        Scale parameter :math:`\rho` corresponding to the average distance of a point
        on the graph to its nearest neighbors.

    Examples
    --------

    .. plot::

        from pycgsp.graph import healpix_nngraph
        from pygsp.plotting import plot_graph
        G, _ = healpix_nngraph(nside=2)
        plot_graph(G)


    Warnings
    --------
    In the newest version of PyGSP (> 0.5.1) the convention is changed: ``Graph.D`` is the divergence operator and
    ``Graph.D.transpose()`` the gradient (see routine `Graph.compute_differential_operator <https://pygsp.readthedocs.io/en/latest/reference/graphs.html#pygsp.graphs.Graph.compute_differential_operator>`_). The code should be adapted when this new version is released.

    """

    npix = hp.nside2npix(nside)
    x, y, z = hp.pix2vec(nside, np.arange(npix))
    R = np.stack((x, y, z), axis=-1)
    cols = hp.get_all_neighbours(nside, np.arange(npix)).transpose().reshape(-1)
    cols[cols == -1] = npix - 1
    rows = np.repeat(np.arange(npix), 8, axis=-1).transpose().reshape(-1)

    # Form sparse affinity matrix from extracted pairs
    W = sp.coo_matrix((cols * 0 + 1, (rows, cols)), shape=(npix, npix))
    # Symmetrize the matrix to obtain an undirected graph.
    extended_row = np.concatenate([W.row, W.col])
    extended_col = np.concatenate([W.col, W.row])
    W.row, W.col = extended_row, extended_col
    W.data = np.concatenate([W.data, W.data])
    W = W.tocsr().tocoo()  # Delete potential duplicate pairs

    # Weight matrix elements according to the exponential kernel
    distance = linalg.norm(R[W.row, :] - R[W.col, :], axis=-1)
    rho = np.mean(distance)
    W.data = np.exp(- (distance / rho) ** 2)
    W = W.tocsc()

    G = _graph_laplacian(W, R, compute_differential_operator=compute_differential_operator,
                         cheb_normalized=cheb_normalized)
    return G, rho


# #############################################################################
# base.py
# =======
# Author : Matthieu Simeoni [matthieu.simeoni@gmail.com]
# #############################################################################

r"""
Base class for polynomial graph operators. Useful for implementing graph convolutions and generalised graph differential operators.
"""
from numbers import Number
from typing import Union

import numpy as np

from pycsou.core.linop import LinearOperator


class PolynomialLinearOperator(LinearOperator):
    r"""
    Polynomial linear operator :math:`P(L)`.

    Given a polynomial :math:`P(x)=\sum_{k=0}^N a_k x^k` and a square linear operator :math:`\mathbf{L}:\mathbb{R}^N\to \mathbb{R}^N,`
    we define the polynomial linear operator :math:`P(\mathbf{L}):\mathbb{R}^N\to \mathbb{R}^N` as:

    .. math::

       P(\mathbf{L})=\sum_{k=0}^N a_k \mathbf{L}^k,

    where :math:`\mathbf{L}^0` is the identity matrix.
    The *adjoint* of :math:`P(\mathbf{L})` is given by:

    .. math::

       P(\mathbf{L})^\ast=\sum_{k=0}^N a_k (\mathbf{L}^\ast)^k.

    Examples
    --------

    .. testsetup::

       import numpy as np

    .. doctest::

       >>> from pycsou.linop import DenseLinearOperator
       >>> from pycgsp.linop.base import PolynomialLinearOperator
       >>> L = DenseLinearOperator(np.arange(64).reshape(8,8))
       >>> PL = PolynomialLinearOperator(LinOp=L, coeffs=[1/2 ,2, 1])
       >>> x = np.arange(8)
       >>> np.allclose(PL(x), x/2 + 2 * L(x) + (L**2)(x))
       True

    """

    def __init__(self, LinOp: LinearOperator, coeffs: Union[np.ndarray, list, tuple]):
        r"""

        Parameters
        ----------
        LinOp: pycsou.core.LinearOperator
            Square linear operator :math:`\mathbf{L}`.
        coeffs: Union[np.ndarray, list, tuple]
            Coefficients :math:`\{a_0,\ldots, a_N\}` of the polynomial :math:`P`.
        """
        self.coeffs = np.asarray(coeffs).astype(LinOp.dtype)
        if LinOp.shape[0] != LinOp.shape[1]:
            raise ValueError('Input linear operator must be square.')
        else:
            self.Linop = LinOp
        super(PolynomialLinearOperator, self).__init__(shape=LinOp.shape, dtype=LinOp.dtype,
                                                       is_explicit=LinOp.is_explicit, is_dense=LinOp.is_dense,
                                                       is_sparse=LinOp.is_sparse,
                                                       is_dask=LinOp.is_dask,
                                                       is_symmetric=LinOp.is_symmetric)

    def __call__(self, x: Union[Number, np.ndarray]) -> Union[Number, np.ndarray]:
        z = x.astype(self.dtype)
        y = self.coeffs[0] * x
        for i in range(1, len(self.coeffs)):
            z = self.Linop(z)
            y += self.coeffs[i] * z
        return y

    def adjoint(self, x: Union[Number, np.ndarray]) -> Union[Number, np.ndarray]:
        if self.is_symmetric:
            return self(x)
        else:
            z = x.astype(self.dtype)
            y = np.conj(self.coeffs[0]) * x
            for i in range(1, len(self.coeffs)):
                z = self.Linop.adjoint(z)
                y += np.conj(self.coeffs[i]) * z
            return y



import configparser
import datetime
import pathlib
import re
from typing import Mapping


def setup_config() -> configparser.ConfigParser:
    """
    Load information contained in `setup.cfg`.
    """
    sphinx_src_dir = pathlib.Path(__file__).parent
    setup_path = sphinx_src_dir / ".." / "setup.cfg"
    setup_path = setup_path.resolve(strict=True)

    with setup_path.open(mode="r") as f:
        cfg = configparser.ConfigParser()
        cfg.read_file(f)
    return cfg


def pkg_info() -> Mapping:
    """
    Load information contained in `PKG-INFO`.
    """
    sphinx_src_dir = pathlib.Path(__file__).parent
    info_path = sphinx_src_dir / ".." / "pycgsp.egg-info" / "PKG-INFO"
    info_path = info_path.resolve(strict=True)

    # Pattern definitions
    pat_version = r"Version: (.+)$"

    with info_path.open(mode="r") as f:
        info = dict(version=None)
        for line in f:
            m = re.match(pat_version, line)
            if m is not None:
                info["version"] = m.group(1)
    return info


# -- Project information -----------------------------------------------------
cfg, info = setup_config(), pkg_info()
project = cfg.get("metadata", "name")
author = cfg.get("metadata", "author")
copyright = f"{datetime.date.today().year}, {author}"
version = release = info["version"]

# -- General configuration ---------------------------------------------------
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.doctest",
    "sphinx.ext.intersphinx",
    "sphinx.ext.mathjax",
    "sphinx.ext.napoleon",
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
    "matplotlib.sphinxext.plot_directive",
]

templates_path = ["_templates"]
master_doc = "index"
exclude_patterns = []
pygments_style = "sphinx"
add_module_names = False
plot_include_source = True

# -- Options for HTML output -------------------------------------------------
html_theme = "sphinx_rtd_theme"
html_theme_options = {"navigation_depth": -1, "titles_only": False}

# -- Options for HTMLHelp output ---------------------------------------------
htmlhelp_basename = "Pycsou-gsp"
html_context = {
    'menu_links_name': 'Repository',
    'menu_links': [
        ('<i class="fa fa-github fa-fw"></i> Source Code', 'https://github.com/matthieumeo/pycsou-gsp'),
        #('<i class="fa fa-users fa-fw"></i> Contributing', 'https://github.com/PyLops/pylops/blob/master/CONTRIBUTING.md'),
    ],
    'doc_path': 'docs/source',
    'github_project': 'matthieumeo',
    'github_repo': 'pycsou-gsp',
    'github_version': 'master',
}

# -- Extension configuration -------------------------------------------------
# -- Options for autosummary extension ---------------------------------------
autosummary_generate = True

# -- Options for autodoc extension -------------------------------------------
autodoc_member_order = "bysource"
autodoc_default_flags = [
    "members",
    # 'inherited-members',
    "show-inheritance",
]
autodoc_inherit_docstrings = True

# -- Options for intersphinx extension ---------------------------------------
intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "NumPy [latest]": ("https://docs.scipy.org/doc/numpy/", None),
    "SciPy [latest]": ("https://docs.scipy.org/doc/scipy/reference", None),
    "pylops [latest]": ("https://pylops.readthedocs.io/en/latest", None),
    "dask [latest]": ("https://docs.dask.org/en/latest/", None),
    "pandas": ("http://pandas.pydata.org/pandas-docs/stable/", None),
    "matplotlib": ("https://matplotlib.org/", None),
    "pycsou": ("https://matthieumeo.github.io/pycsou/html/", None)
}

# -- Options for napoleon extension ------------------------------------------
napoleon_google_docstring = False
napoleon_numpy_docstring = True

# -- Options for todo extension ----------------------------------------------
todo_include_todos = True



#!/usr/bin/env python3

# #############################################################################
# test.py
# =======
# Author : Matthieu Simeoni [matthieu.simeoni@gmail.com]
# #############################################################################

"""
Run tests.
"""

import argparse
import pathlib
import subprocess
import sys

project_root_dir = pathlib.Path(__file__).parent.absolute()
cmds = dict(
    doctest=[f'sphinx-build -b doctest "{project_root_dir}/doc" "{project_root_dir}/build/doctest"']
)

parser = argparse.ArgumentParser(
    description="pycsou-gsp test runner.",
    epilog="When run with no arguments, all tests are executed.",
)
parser.add_argument("-e", help="Name of test to run.", type=str, choices=cmds.keys())
args = parser.parse_args()


def run_test(test_name):
    """
    Execute a test on the command line.

    Parameters
    ----------
    test_name : str
        Name of a key in `cmds`.
    """
    if not isinstance(test_name, str):
        raise ValueError("Parameter[test_name] must be a str.")
    if test_name not in cmds:
        raise ValueError("Parameter[test_name] is not a valid test.")

    status = subprocess.run(
        "; ".join(cmds[test_name]),
        stdin=None,
        stdout=sys.stdout,
        stderr=sys.stderr,
        shell=True,
        cwd=project_root_dir,
    )
    return status


if __name__ == "__main__":
    status = []

    if args.e is None:
        for test in cmds:
            s = run_test(test)
            status.append(s)
    else:
        s = run_test(args.e)
        status.append(s)

    print("\nSummary\n=======")
    for s in status:
        print("Success" if (s.returncode == 0) else "Failure")
        cmd_list = s.args.split("; ")
        for c in cmd_list:
            print(f"   {c}")



# #############################################################################
# diff.py
# =======
# Author : Matthieu Simeoni [matthieu.simeoni@gmail.com]
# #############################################################################

r"""
Graph differential operators.
"""

import numpy as np
import pygsp

from pycsou.core.linop import LinearOperator
from pycsou.linop.base import SparseLinearOperator, IdentityOperator
from pycgsp.linop.base import PolynomialLinearOperator


class GraphLaplacian(LinearOperator):
    r"""
    Graph Laplacian.

    Normalised graph Laplacian for signals defined on graphs.

    Examples
    --------

    .. plot::

       import numpy as np
       from pygsp.graphs import Ring
       from pycgsp.linop.diff import GraphLaplacian
       np.random.seed(1)
       G = Ring(N=32, k=4)
       G.compute_laplacian(lap_type='normalized')
       G.set_coordinates(kind='spring')
       x = np.arange(G.N)
       signal = np.piecewise(x, [x < G.N//3, (x >= G.N//3) * (x< 2 * G.N//3), x>=2 * G.N//3], [lambda x: -x, lambda x: 3 * x - 4 * G.N//3, lambda x: -0.5 * x + G.N])
       Lap = GraphLaplacian(Graph=G)
       lap_sig = Lap * signal
       plt.figure()
       ax=plt.gca()
       G.plot_signal(signal, ax=ax, backend='matplotlib')
       plt.title('Signal')
       plt.axis('equal')
       plt.figure()
       plt.plot(signal)
       plt.title('Signal')
       plt.figure()
       ax=plt.gca()
       G.plot_signal(lap_sig, ax=ax, backend='matplotlib')
       plt.title('Laplacian of signal')
       plt.axis('equal')
       plt.figure()
       plt.plot(-lap_sig)
       plt.title('Laplacian of signal')

    Notes
    -----
    For undirected graphs, the normalized graph Laplacian is defined as

    .. math:: \mathbf{L} = \mathbf{I} - \mathbf{D}^{-1/2} \mathbf{W} \mathbf{D}^{-1/2},

    where :math:`\mathbf{I}` is the identity matrix, :math:`\mathbf{W}` is the weighted adjacency matrix and :math:`\mathbf{D}` the
    weighted degree matrix.

    For directed graphs, the Laplacians are built from a symmetrized
    version of the weighted adjacency matrix that is the average of the
    weighted adjacency matrix and its transpose. As the Laplacian is
    defined as the divergence of the gradient, it is not affected by the
    orientation of the edges.

    For both Laplacians, the diagonal entries corresponding to disconnected
    nodes (i.e., nodes with degree zero) are set to zero.

    The ``GraphLaplacian`` operator is self-adjoint.

    See Also
    --------
    :py:class:`~pycgsp.linop.diff.GraphGradient`, :py:func:`~pycgsp.linop.diff.GeneralisedGraphLaplacian`
    :py:class:`~pycgsp.linop.conv.GraphConvolution`

    """

    def __init__(self, Graph: pygsp.graphs.Graph, dtype: type = np.float):
        r"""
        Parameters
        ----------
        Graph: `pygsp.graphs.Graph <https://pygsp.readthedocs.io/en/stable/reference/graphs.html#pygsp.graphs.Graph>`_
            Graph on which the signal is defined, with normalised Laplacian ``Graph.L`` precomputed (see `pygsp.graphs.Graph.compute_laplacian(lap_type='normalized') <https://pygsp.readthedocs.io/en/stable/reference/graphs.html#pygsp.graphs.Graph.compute_laplacian>`_.
        dtype: type
            Type of the entries of the graph filer.

        Raises
        ------
        AttributeError
            If ``Graph.L`` does not exist.
        NotImplementedError
            If ``Graph.lap_type`` is 'combinatorial'.

        """
        self.Graph = Graph
        if Graph.L is None:
            raise AttributeError(
                r'Please compute the normalised Laplacian of the graph with the routine https://pygsp.readthedocs.io/en/stable/reference/graphs.html#pygsp.graphs.Graph.compute_laplacian')
        elif Graph.lap_type != 'normalized':
            raise NotImplementedError(r'Combinatorial graph Laplacians are not supported.')
        else:
            self.L = self.Graph.L.tocsc()
        super(GraphLaplacian, self).__init__(shape=self.Graph.W.shape, dtype=dtype, is_explicit=False,
                                             is_dense=False, is_sparse=False, is_dask=False,
                                             is_symmetric=True)

    def __call__(self, x: np.ndarray) -> np.ndarray:
        return self.L.dot(x)

    def adjoint(self, y: np.ndarray) -> np.ndarray:
        return self(y)


class GraphGradient(LinearOperator):
    r"""
    Graph gradient.

    Gradient operator for signals defined on graphs.

    Examples
    --------

    .. testsetup::

       import numpy as np
       from pygsp.graphs import Ring
       from pycgsp.linop.diff import GraphLaplacian, GraphGradient
       np.random.seed(1)

    .. doctest::

       >>> G = Ring(N=32, k=4)
       >>> G.compute_laplacian(lap_type='normalized')
       >>> G.compute_differential_operator()
       >>> G.set_coordinates(kind='spring')
       >>> x = np.arange(G.N)
       >>> signal = np.piecewise(x, [x < G.N//3, (x >= G.N//3) * (x< 2 * G.N//3), x>=2 * G.N//3], [lambda x: -x, lambda x: 3 * x - 4 * G.N//3, lambda x: -0.5 * x + G.N])
       >>> Lap = GraphLaplacian(Graph=G)
       >>> Grad = GraphGradient(Graph=G)
       >>> lap_sig = Lap * signal
       >>> lap_sig2 = Grad.adjoint(Grad(signal))
       >>> np.allclose(lap_sig, lap_sig2)
       True

    Notes
    -----
    The adjoint of the ``GraphGradient`` operator is called the graph divergence operator.

    Warnings
    --------
    In the newest version of PyGSP (> 0.5.1) the convention is changed: ``Graph.D`` is the divergence operator and
    ``Graph.D.transpose()`` the gradient (see routine `Graph.compute_differential_operator <https://pygsp.readthedocs.io/en/latest/reference/graphs.html#pygsp.graphs.Graph.compute_differential_operator>`_). The code should be adapted when this new version is released.

    See Also
    --------
    :py:class:`~pycgsp.linop.diff.GraphLaplacian`, :py:func:`~pycgsp.linop.diff.GeneralisedGraphLaplacian`

    """

    def __init__(self, Graph: pygsp.graphs.Graph, dtype: type = np.float):
        r"""
        Parameters
        ----------
        Graph: `pygsp.graphs.Graph <https://pygsp.readthedocs.io/en/stable/reference/graphs.html#pygsp.graphs.Graph>`_
            Graph on which the signal is defined, with differential operator ``Graph.D`` precomputed (see `pygsp.graphs.Graph.compute_differential_operator() <https://pygsp.readthedocs.io/en/stable/reference/graphs.html#pygsp.graphs.Graph.compute_differential_operator>`_.
        dtype: type
            Type of the entries of the graph filer.

        Raises
        ------
        AttributeError
            If ``Graph.D`` does not exist.
        """
        self.Graph = Graph
        if Graph.D is None:
            raise AttributeError(
                r'Please compute the differential operator of the graph with the routine https://pygsp.readthedocs.io/en/stable/reference/graphs.html#pygsp.graphs.Graph.compute_differential_operator')
        else:
            self.D = self.Graph.D.tocsc()
        super(GraphGradient, self).__init__(shape=self.Graph.W.shape, dtype=dtype, is_explicit=False,
                                            is_dense=False, is_sparse=False, is_dask=False,
                                            is_symmetric=True)

    def __call__(self, x: np.ndarray) -> np.ndarray:
        return self.D.dot(x)

    def adjoint(self, y: np.ndarray) -> np.ndarray:
        return self.D.conj().transpose().dot(y)


def GeneralisedGraphLaplacian(Graph: pygsp.graphs.Graph, kind: str = 'iterated', **kwargs):
    r"""
    Generalised graph Laplacian operator.

    Generalised Laplacian operator signals defined on graphs.

    Parameters
    ----------
    Graph: `pygsp.graphs.Graph <https://pygsp.readthedocs.io/en/stable/reference/graphs.html#pygsp.graphs.Graph>`_
        Graph on which the signal is defined, with normalised Laplacian ``Graph.L`` precomputed (see `pygsp.graphs.Graph.compute_laplacian(lap_type='normalized') <https://pygsp.readthedocs.io/en/stable/reference/graphs.html#pygsp.graphs.Graph.compute_laplacian>`_.
    dtype: type
        Type of the entries of the graph filer.
    kind: str
        Type of generalised differential operator (``'iterated'``, ``'sobolev'``, ``'polynomial'``).
        Depending on the cases, the ``GeneralisedLaplacian`` operator is defined as follows:

        * ``'iterated'``: :math:`\mathscr{D}=\mathbf{L}^N`,
        * ``'sobolev'``: :math:`\mathscr{D}=(\alpha^2 \mathrm{Id}-\mathbf{L})^N`, with :math:`\alpha\in\mathbb{R}`,
        * ``'polynomial'``: :math:`\mathscr{D}=\sum_{n=0}^N \alpha_n \mathbf{L}^n`,  with :math:`\{\alpha_0,\ldots,\alpha_N\} \subset\mathbb{R}`,

        where :math:`\mathbf{L}` is the :py:class:`~pycgsp.linop.diff.GraphLaplacian` operator.
    kwargs: Any
        Additional arguments depending on the value of ``kind``:

        * ``'iterated'``: ``kwargs={order: int}`` where ``order`` defines the exponent :math:`N`.
        * ``'sobolev'``: ``kwargs={order: int, constant: float}`` where ``order`` defines the exponent :math:`N` and ``constant`` the scalar :math:`\alpha\in\mathbb{R}`.
        * ``'polynomial'``: ``kwargs={coeffs: Union[np.ndarray, list, tuple]}`` where ``coeffs`` is an array containing the coefficients :math:`\{\alpha_0,\ldots,\alpha_N\} \subset\mathbb{R}`.

    Raises
    ------
    AttributeError
        If ``Graph.L`` does not exist.
    NotImplementedError
        If ``Graph.lap_type`` is 'combinatorial'.
    NotImplementedError
        If ``kind`` is not 'iterated', 'sobolev' or 'polynomial'.

    Examples
    --------

    .. plot::

       import numpy as np
       from pygsp.graphs import Ring
       from pycgsp.linop.diff import GeneralisedGraphLaplacian
       np.random.seed(1)
       G = Ring(N=32, k=4)
       G.compute_laplacian(lap_type='normalized')
       G.set_coordinates(kind='spring')
       x = np.arange(G.N)
       signal = np.piecewise(x, [x < G.N//3, (x >= G.N//3) * (x< 2 * G.N//3), x>=2 * G.N//3], [lambda x: -x, lambda x: 3 * x - 4 * G.N//3, lambda x: -0.5 * x + G.N])
       Dop = GeneralisedGraphLaplacian(Graph=G, kind='polynomial', coeffs=[1,-1,2])
       gen_lap = Dop * signal
       plt.figure()
       ax=plt.gca()
       G.plot_signal(signal, ax=ax, backend='matplotlib')
       plt.title('Signal')
       plt.axis('equal')
       plt.figure()
       ax=plt.gca()
       G.plot_signal(gen_lap, ax=ax, backend='matplotlib')
       plt.title('Generalized Laplacian of signal')
       plt.axis('equal')


    Notes
    -----
    The ``GeneralisedGraphLaplacian`` operator is self-adjoint.

    See Also
    --------
    :py:class:`~pycgsp.linop.diff.GraphLaplacian`


    """
    if Graph.L is None:
        raise AttributeError(
            r'Please compute the normalised Laplacian of the graph with the routine https://pygsp.readthedocs.io/en/stable/reference/graphs.html#pygsp.graphs.Graph.compute_laplacian')
    elif Graph.lap_type != 'normalized':
        raise NotImplementedError(r'Combinatorial graph Laplacians are not supported.')
    else:
        L = Graph.L.tocsc()
        LapOp = SparseLinearOperator(L, is_symmetric=True)

    if kind == 'iterated':
        N = kwargs['order']
        Dgen = LapOp ** N
    elif kind == 'sobolev':
        I = IdentityOperator(size=LapOp.shape[0] * LapOp.shape[1])
        alpha = kwargs['constant']
        N = kwargs['order']
        Dgen = ((alpha ** 2) * I - LapOp) ** N
    elif kind == 'polynomial':
        coeffs = kwargs['coeffs']
        Dgen = PolynomialLinearOperator(LinOp=LapOp, coeffs=coeffs)
    else:
        raise NotImplementedError(
            'Supported generalised Laplacian types are: iterated, sobolev, polynomial.')
    return Dgen



# #############################################################################
# conv.py
# =======
# Author : Matthieu Simeoni [matthieu.simeoni@gmail.com]
# #############################################################################

r"""
Graph convolution operator.
"""
from typing import Union

import numpy as np
import pygsp

from pycsou.linop.base import SparseLinearOperator
from pycgsp.linop.base import PolynomialLinearOperator


class GraphConvolution(PolynomialLinearOperator):
    r"""
    Graph convolution.

    Convolve a signal :math:`\mathbf{u}\in\mathbb{C}^N` defined on a graph with a polynomial filter :math:`\mathbf{D}:\mathbb{C}^N\rightarrow \mathbb{C}^N`
    of the form:

    .. math::

       \mathbf{D}=\sum_{k=0}^K \theta_k \mathbf{L}^k,

    where :math:`\mathbf{L}:\mathbb{C}^N\rightarrow \mathbb{C}^N` is the *normalised graph Laplacian* (see [FuncSphere]_ Section 2.3 of Chapter 6).

    Examples
    --------

    .. testsetup::

       import numpy as np
       from pygsp.graphs import RandomRegular
       from pycgsp.linop.conv import GraphConvolution
       np.random.seed(0)

    .. doctest::

       >>> G = RandomRegular(seed=0)
       >>> G.compute_laplacian(lap_type='normalized')
       >>> signal = np.random.binomial(n=1,p=0.2,size=G.N)
       >>> coefficients = np.ones(shape=(3,))
       >>> ConvOp = GraphConvolution(Graph=G, coefficients=coefficients)
       >>> filtered = ConvOp * signal

    .. plot::

       import numpy as np
       from pygsp.graphs import Ring
       from pycgsp.linop.conv import GraphConvolution
       np.random.seed(0)
       G = Ring(N=32, k=2)
       G.compute_laplacian(lap_type='normalized')
       G.set_coordinates(kind='spring')
       signal = np.random.binomial(n=1,p=0.2,size=G.N)
       coefficients = np.ones(3)
       ConvOp = GraphConvolution(Graph=G, coefficients=coefficients)
       e1 = np.zeros(shape=G.N)
       e1[0] = 1
       filter = ConvOp * e1
       filtered = ConvOp * signal
       plt.figure()
       ax=plt.gca()
       G.plot_signal(signal, ax=ax, backend='matplotlib')
       plt.title('Signal')
       plt.axis('equal')
       plt.figure()
       ax=plt.gca()
       G.plot_signal(filter, ax=ax, backend='matplotlib')
       plt.title('Filter')
       plt.axis('equal')
       plt.figure()
       ax=plt.gca()
       G.plot_signal(filtered, ax=ax, backend='matplotlib')
       plt.title('Filtered Signal')
       plt.axis('equal')

    Notes
    -----
    The ``GraphConvolution`` operator is self-adjoint and operates in a matrix-free fashion, as described in Section 4.3, Chapter 7 of  [FuncSphere]_.

    See Also
    --------
    :py:class:`~pycgsp.linop.diff.GraphLaplacian`

    """

    def __init__(self, Graph: pygsp.graphs.Graph, coefficients: Union[np.ndarray, list, tuple]):
        r"""
        Parameters
        ----------
        Graph: `pygsp.graphs.Graph <https://pygsp.readthedocs.io/en/stable/reference/graphs.html#pygsp.graphs.Graph>`_
            Graph on which the signal is defined, with normalised Laplacian ``Graph.L`` precomputed (see `pygsp.graphs.Graph.compute_laplacian(lap_type='normalized') <https://pygsp.readthedocs.io/en/stable/reference/graphs.html#pygsp.graphs.Graph.compute_laplacian>`_.
        coefficients: Union[np.ndarray, list, tuple]
            Coefficients :math:`\{\theta_k, \,k=0,\ldots,K\}\subset \mathbb{C}` of the polynomial filter.
        dtype: type
            Type of the entries of the graph filer.

        Raises
        ------
        AttributeError
            If ``Graph.L`` does not exist.
        NotImplementedError
            If ``Graph.lap_type`` is 'combinatorial'.
        """
        self.Graph = Graph
        if Graph.L is None:
            raise AttributeError(
                r'Please compute the normalised Laplacian of the graph with the routine https://pygsp.readthedocs.io/en/stable/reference/graphs.html#pygsp.graphs.Graph.compute_laplacian')
        elif Graph.lap_type != 'normalized':
            raise NotImplementedError(r'Combinatorial graph Laplacians are not supported.')
        else:
            L = self.Graph.L.tocsc()
            Lop = SparseLinearOperator(L, is_symmetric=True)
        self.coefficients = coefficients
        super(GraphConvolution, self).__init__(LinOp=Lop, coeffs=self.coefficients)
